/* pfdp1.c */

/* This program implements a parallel version of the forward dynamic 
   programming algorithm presented in "Power Generation, Operation, and 
   Control" by A. J. Wood and B. F. Wollenberg.  This program follows 
   the  flowchart in figure 5.5, except that the initialization has been 
   worked into the first time through the loop.  This program permits 
   different values of x and n, and requires a sychronization and 
   sequential calculation between each stage.  This particular program 
   is a specialized version to explicitly work with the diagram in 
   FIG 5.17 on page 140 and problem 5.1 on page 147.  This program is 
   the first proof-of-concept of the parallelized forward dynamic 
   programming algorithm for unit commitment.  The forward dynamic 
   programming algorithm is correct; however, data will be explicitly 
   entered from files for each state in each period rather than calculating
   the feasible states from unit priority data and load patterns. */

#include <stdio.h>
#include <string.h>
#include <cthreads.h>

/* define some constants */
#define MAXPERIODS    5
#define MAXNODES      4
#define MAXNAMELENGTH 5
#define FILENAME_MAX  24
#define YES           1
#define NO            0
#define MAYBE         1
#define INFINITY      999

/* define a structure that represents a state or node.  a node is
   completely defined by five items: 
      1. name (name)
      2. least total cost to arrive at this node (fcost)
      3. production cost for this node (pcost)
      4. transition costs from the nodes in the previous period
         to this node (scost[]) 
      5. pointer to the state in the previous period that is
          potentially in the optimal schedule (backpoint) */

struct node {char name[MAXNAMELENGTH];
             int  fcost;
             int  pcost;
             int  scost[MAXNODES];
             int  backpoint};

/* define a matrix of nodes for various periods.  The 0th period is the
   initial state, while there are (m) periods to be examined.  Consequently,
   there are a total of (m+1) periods.  At most (x) nodes will be 
   examined each period. */

struct node nodes[MAXNODES][MAXPERIODS+1];

/* define an array to hold the (n) feasible states in the interval 
   previous to the one currently being examined.  These states will 
   be used in further calculations. */

int    l[MAXNODES];

/* define a structure that matches cost to arrive at a state with a
   store of the index to that value */

struct cost_index {int fcost;
                   int i};

/* define an array of cost-index structures */

struct cost_index incost[MAXNODES];

int n;  /* number of lowest cost strategies to save */
int x;  /* number of states to be examined per period */
int m;  /* number of periods to be examined */

/* define user-supplied functions */
/* function names are designed to be self-explainatory */

void init_mach_local();
void start_sync();
void end_sync();
void start_resume();
void end_resume();
void init_constants();
void init_nodes();
void print_nodenames();
void print_scost();
void calc_l();
void sort_array();
void calc_fcost();
void trace_optimal_schedule();

/* define an accumulator variable 'COUNT.count' that will identify when 
   all x nodes have been calculated.  it is protected from simultaneous 
   access by the mutex COUNT.MUTEX.  if (COUNT.count < x) -> some 
   thread has yet to finish calculations.  if (COUNT.count == x) -> it is 
   time to calculate, the array of feasible states from this stage */

struct count_obj {int              count;
                  struct mutex     MUTEX;
                  struct condition CONDITION} COUNT;

/* define a structure that includes an indicator variable to identify 
   when to resume processing of the next stage for the corresponding node 
   or processor if (RESUME.resume == 0) -> the nodes in the (k+1)st 
   stage are ready to calculate.  if (RESUME.resume == 1) -> the 
   synchronization has yet to occur, so the threads must wait */

struct resume_obj {int              resume;
                   struct mutex     MUTEX;
                   struct condition CONDITION} RESUME;

/* This program is a parallelized version of the program FDP.C, or the 
   forward dynamic programming example. Operations in the main routine 
   are abstracted by extensive use of functions. */

/* the main program is called with no arguments */

main()
{
 int k;   /* variable that represents the period under examination */
 int q;   /* an index variable */
 int processor_num[MAXNODES];
 int *pq;
 struct t_id {int *thread_id} THREAD[MAXNODES]; /* an array of thread ids 
                                                   to be used when joining 
                                                   the threads */

 init_mach_local();                /* initialize all mach routine */
 init_constants();           /* get constants from the user */
 init_nodes();               /* initialize all nodes from a file */

 for (k = 1; k <= m; k++)  /* calculate fcost for all the periods */
 {
printf("period |%d| \n",k);
fflush(stdout);
  calc_l( (k-1) );         /* identify the n most feasible states in 
                              the previous period */

  if (k == 1)              /* fork the processes to calculate fcost for
                              all the periods */
  {
   for (q = 0;q < x;q++)
   {
printf("fork process |%d| \n",q);
fflush(stdout);
    processor_num[q] = q;
    pq = processor_num + q;
/* fork the threads and store the thread id */
    THREAD[q].thread_id = cthread_fork(calc_fcost, pq);
   }
  }
 }
 for (q = 0;q < x;q++)
 {
printf("join process |%d| \n",q);
fflush(stdout);
  cthread_join(THREAD[q].thread_id);   /* join all the threads */
 }
 trace_optimal_schedule(m);     /* trace the optimal schedule */
}


/* This function initializes all mach specific variables */

void init_mach_local()
{
 cthread_init();  /* initialize cthreads functions */

/* allocate all mutexs and conditionals */
 mutex_init(&COUNT.MUTEX);
 condition_init(&COUNT.CONDITION);

 mutex_init(&RESUME.MUTEX);
 condition_init(&RESUME.CONDITION);
}


/* This function sets up the condition to wait until all values of fcost 
   have been calculated at a stage */

void start_sync()
{
 mutex_lock(&COUNT.MUTEX);            /* lock the mutex */
printf("enter start-sync \n");
fflush(stdout);

/* check to see if all nodes have been calculated */
 while(COUNT.count < x)
 {
/* wait for the CONDITION condition to be signalled */
  condition_wait(&COUNT.CONDITION,&COUNT.MUTEX);  
 }
 mutex_unlock(&COUNT.MUTEX);          /* unlock the mutex */
printf("exit start-sync \n");
fflush(stdout);

}


/* This function resets the value of count and the indicator variable, 
   RESUME.resume */

void end_sync()
{
printf("enter end-sync \n");
fflush(stdout);

 mutex_lock(&COUNT.MUTEX);        /* lock the mutex */
 COUNT.count = 0;                      /* reset count to zero */
 mutex_unlock(&COUNT.MUTEX);      /* unlock the mutex */

 mutex_lock(&RESUME.MUTEX)    /* lock the mutex */
 RESUME.resume = 0;              /* reset resume to zero */
/* signal to resume the threads */
 condition_broadcast(&RESUME.CONDITION);
 mutex_unlock(&RESUME.MUTEX);  /* unlock the mutex */
printf("exit end-sync \n");
fflush(stdout);

}


/* This function sets up the condition to wait until the array of feasible
   states from the previous stage has been calculated */

void start_resume()
{
printf("enter start-resume \n");
fflush(stdout);

 mutex_lock(&RESUME.MUTEX);       /* lock the mutex */
 while(RESUME.resume >= x)              /* check to see if the qth value of */
                                    /* fcost has been calculated */
 {
/* wait for the RESUME.CONDITION condition to be signalled */
  condition_wait(&RESUME.CONDITION,&RESUME.MUTEX);

 }
 mutex_unlock(&RESUME.MUTEX);     /* unlock the mutex */
printf("exit start_resume \n");
fflush(stdout);

}


/* This function signals that a value of fcost has been calculated */

void end_resume()
 {
printf("enter end-resume \n");
fflush(stdout);

 mutex_lock(&COUNT.MUTEX);            /* lock the mutex */
 COUNT.count = COUNT.count + 1;            /* increment count */
 condition_signal(&COUNT.CONDITION);  /* signal the SYNC condition */
 mutex_unlock(&COUNT.MUTEX);          /* unlock the mutex */

 mutex_lock(&RESUME.MUTEX);        /* lock the mutex */
/* indicate the value of fcost has been calculated */
 RESUME.resume += 1;
 mutex_unlock(&RESUME.MUTEX);      /* unlock the mutex */
printf("exit end-resume \n");
fflush(stdout);

 }


/* This function queries the user for input for constants in the program */

void init_constants()
 {
  printf("input the number of periods to be examined (m) ");
  scanf ("%d",&m);
/* the number of periods cannot be greater than MAXPERIODS */
  if (m > MAXPERIODS)
   {
    m = MAXPERIODS;
    printf("the number of periods cannot be greater than %d \n",MAXPERIODS);
   }

  printf("input the number of states to be examined (x) ");
  scanf ("%d",&x);
/* the number of states cannot be greater than MAXNODES */
  if (x > MAXNODES)
   {
    x = MAXNODES;
    printf("the number of states cannot be greater than %d \n",MAXNODES);
   }

  printf("input the number of strategies to be saved (n) ");
  scanf ("%d",&n);
/* the number of strategies saved cannot be greater than MAXNODES */
  if (n > MAXNODES)
   {
    n = MAXNODES;
    printf("the number of strategies saved cannot be greater than %d \n",
                                                      MAXNODES);
   }
 }


/* This function reads a file to enter the node names and transition
   cost data.  The data is stored with identifier fields for the period
   and the number of the node in the period.  The matrix of nodes for
   the various periods need not be complete.  The only requirements are
   that the user correctly generate the transition costs, and that there
   is at least one node per period.  This input routine was specifically
   designed to handle data similar to FIG. 5.17 or Problem 5.1 in "Power
   Generation, Operation, and Control".

   The routine intializes the matrix of structures before data is entered,
   so that non-assigned states have '*' for a name and the cost to arrive
   at a state and all transition costs are set to INFINITY, a sufficiently
   large integer.  Also for this particular input data, production costs
   are assumed to be zero, so all production values in the matrix of
   structures are assigned to 0.

   The input format, with data ranges, is

           period | node# | node-name | array of scost

   where

   period is an integer and takes values [0-MAXPERIODS]. 
   node# is an integer and takes values [0-(MAXNODES-1)].
   node-name is a character with length less than MAXNAMELENGTH.
   scost values are in an integer array with MAXNODES number of values.  If
   no link exists with a previous node, that value is sent to -1, and the
   program will assign that transition cost to INFINITY.  My appoligies
   for hardwiring the array of scost values into the program, but this is 
   a proof-of-concept program and later programs will use a different 
   scheme to set up node information, when the power generation unit 
   commitment problem is addressed. */

void init_nodes()
 {
  int  i,j,k,q;                /* index variables */
  char filename[FILENAME_MAX]; /* input filename */
  FILE *fp;                    /* pointer to the file stream */
  int  period;                 /* input variable for the period */
  int  nodenum;                /* input variable for the node number */
  char name[MAXNAMELENGTH];    /* input variable for the node name */
/* these scost input variables assume MAXNODES = 4 */
  int  scost0;                 /* input variable for 1st scost */
  int  scost1;                 /* input variable for 2nd scost */
  int  scost2;                 /* input variable for 3rd scost */
  int  scost3;                 /* input variable for 4th scost */

/* initialize the values at each node in the matrix of nodes */
  for (i=0;i<MAXNODES;i++)
   {
    for (j=0;j<=MAXPERIODS;j++)
     {
      nodes[i][j].name[0] = '*';         /* set the node name to a dash */
      nodes[i][j].name[1] = '\0';        /* set the end-of-string marker */
      nodes[i][j].fcost   = INFINITY;    /* set the total cost to INFINITY */
      nodes[i][j].pcost   = 0;           /* set the production cost to 0 */
      for (k=0;k<MAXNODES;k++)
       {
        nodes[i][j].scost[k] = INFINITY; /* set the transition costs from */
       }                                 /* nodes in the previous period */
                                         /* to INFINITY */
      nodes[i][j].backpoint = -1;        /* set the backpointer to an */
                                         /* impossible value */
     }
   }

/* initialize the costs in the 0th state to 0 */
  for (i=0;i<MAXNODES;i++)
   {
    nodes[i][0].fcost = 0;
   }

/* open the node data file */
  printf("input the filename for the node data ");
  scanf ("%s",filename);
  printf("\n");

  fp = fopen(filename, "r"); 
/* read the node data file until an EOF is encountered */

  while ((fscanf(fp,"%d %d %s %d %d %d %d", &period,
                                            &nodenum,
                                            name,
                                            &scost0,
                                            &scost1,
                                            &scost2,
                                            &scost3)) != EOF)
/* copy the input data to the data structure */
   {
    strncpy(nodes[nodenum][period].name,name,3);
    if (scost0 != -1) nodes[nodenum][period].scost[0] = scost0;
    if (scost1 != -1) nodes[nodenum][period].scost[1] = scost1;
    if (scost2 != -1) nodes[nodenum][period].scost[2] = scost2;
    if (scost3 != -1) nodes[nodenum][period].scost[3] = scost3;
   }

  fclose(fp);

/* echo the data to the screen */
  print_nodenames();
  print_scost();

/* initialize the shared indicator variables */
  RESUME.resume = 1;
  COUNT.count = 4;
 }


/* This function prints the name and least total cost to arrive at the
   node. */

void print_nodenames()
 {
  int  i,j;  /* index variables */

  printf("\n");
  printf("NODENAME/COST MATRIX \n");

  for (i=0;i<MAXNODES;i++)
   {
    for (j=0;j<=MAXPERIODS;j++)
     {
/* name - cost are represented as a tuple */
      printf("(%2s - %4d)  ",nodes[i][j].name,
                          nodes[i][j].fcost);
     }
    printf("\n");
   }
 }


/* This function prints the transition costs to arrive at a node. */

void print_scost()
 {
  int  i,j,k;  /* index variables */

  printf("\n");
  printf("TRANSITION COST MATRIX \n");

  for (i=0;i<MAXNODES;i++)
   {
    for (k=0;k<MAXNODES;k++)
     {
      for (j=0;j<=MAXPERIODS;j++)
       {
        printf("%4d",nodes[i][j].scost[k]);
       }
      printf("\n");
     }
    printf("\n");
   }
 }


/* This function calculates the n feasible states in the previous
   period that will be used in further optimal unit commitment
   calculations. */

void calc_l(k)
  int k; /* period under examination */
 {
  int i,j;  /* index variables */
  start_sync();  /* synchronize to access 'count' */

/* copy the costs from the data structure for this time period to the
   working array */
  for (i=0;i<x;i++)
   {
    incost[i].fcost = nodes[i][k].fcost;
    incost[i].i     = i;
   }
  sort_array(); /* sort the working array of costs */

/* generate l[], the array of the n most promising states in the 
   previous period by reading the stored indices */

  for (i=0;i<x;i++)
   {
    l[i] = incost[i].i;
   }

  end_sync();  /* notify the threads to resume processing */
 }


/* This function is an application of the simple selection sort from 
   R. Sedgewick, "Algorithms". A selection sort was chosen because 
   the number of items to be sorted will usually be small (< 10). */

void sort_array() /* array to be sorted is incost[] */
 {
  int i,j;   /* index variables */
  int min;   /* variable to hold the location of the minimum value */
  int tc;    /* variable to hold the minimum value of fcost for each 
                search cycle */
  int ti;    /* variable to hold the value of i corresponding to fcost
                for each search cycle */

  for (i=0;i<MAXNODES-1;i++)
   {
    min = i;

    for (j=i+1;j<MAXNODES;j++)
     {
      if (incost[j].fcost < incost[min].fcost) min = j;
     }

    tc = incost[min].fcost;
    ti = incost[min].i;
    incost[min].fcost = incost[i].fcost;
    incost[min].i     = incost[i].i;
    incost[i].fcost   = tc;
    incost[i].i       = ti;
   }
 }


/* This function calculates the minimum cost to arrive at each node
   in a particular period. */

void calc_fcost(pq)
  int *pq;  /* thread identifier*/
 {
  int q = *pq;
  int j,k;    /* index variables */
  int temp;   /* temporary storage value */

  for (k = 1;k <= m;k++)
   {
    start_resume();  /* wait until sequential processing completed */

/* calculate the minimum value of fcost */

    nodes[q][k].fcost = nodes[q][k].pcost +
                        nodes[q][k].scost[ l[0] ] +
                        nodes[ l[0] ][k-1].fcost;
    nodes[q][k].backpoint = l[0];
    for (j=1;j<n;j++)
     {
      temp = nodes[q][k].pcost + 
             nodes[q][k].scost[ l[j] ] + 
             nodes[ l[j] ][k-1].fcost;
      if (temp < nodes[q][k].fcost)
       {
        nodes[q][k].fcost = temp;
        nodes[q][k].backpoint = l[j];
       }
     }
    end_resume();  /* signal that processing has completed */
   }
 }


/* This function traces the optimal schedule by starting at the least
   cost final node, and works backward until the 0th state is reached, or
   until a situation is reached where no optimal schedule can be traced.
   This situation could occur when a critical node is discarded.  If
   such a situation would occur, the problem would require reprocessing
   with a larger value of n, or the program could be augmented to include
   a roll-back option that limits the amount of recalculations. */

void trace_optimal_schedule(m)
  int m; /* number of periods to be examined */
 {
  int i,j,k,p;     /* index variables */
  int temp;        /* value used to find the minimum cost final state */
  int test;        /* indicator variables */
  int isave;       /* variable to save the period where the trace failed */
  int os[MAXPERIODS+1]; /* array of nodes in the optimal schedule */

/* echo the data to the screen */
  print_nodenames();

/* find the minimum cost node in the last period */
  p = 0;
  temp = nodes[0][m].fcost;
  for (j=1;j<MAXNODES;j++)
   {
    if (nodes[j][m].fcost < temp)
     {
      p = j;
      temp = nodes[j][m].fcost;
     }
   }
/* record the location of the minimum cost node in the last period */
  os[m] = p;

/* trace through the remaining periods */
/* if test is set to NO, then an optimal schedule cannot be found */
  test = MAYBE;
  for (i=m-1;(i>=0 && test==MAYBE);i--)
   {
    os[i] = nodes[ os[i+1] ][i+1].backpoint;
/* check to see if this was the value used in calculating the optimal
   schedule */
    temp = (nodes[ os[i] ][i].fcost +
            nodes[ os[i+1] ][i+1].pcost +
            nodes[ os[i+1] ][i+1].scost[ os[i] ]);
    if ( os[i] == -1 ||
         nodes[ os[i+1] ][i+1].fcost != temp)
     {
      test = NO;
/* save the period where the trace failed */
      isave = i;
     }
   }

  printf("\n");

/* print information on the trace */
  if (test == YES)
/* an optimal schedule was found */
   {
    printf("OPTIMAL SCHEDULE\n\n");

    for (i=0;i<=m;i++)
     {
      printf("period #%d  (%s - %4d)\n",i,nodes[ os[i] ][i].name,
                                     nodes[ os[i] ][i].fcost);
     }
   }
  else
/* an optimal schedule was not found */
   {
    printf("TRACE WAS UNABLE TO FIND AN OPTIMAL SCHEDULE\n");
    printf("TRACE FAILED IN PERIOD - %d");

    printf("PARTIAL OPTIMAL SCHEDULE\n");
    for (i=isave+1;i<=m;i++)
     {
      printf("period #%d  (%s - %4d)\n",i,nodes[ os[i] ][i].name,
                                     nodes[ os[i] ][i].fcost);
     }
   }
 }

